# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import pkgutil
from collections.abc import Iterable
from typing import Any, Literal

import h5py
import numpy as np

import bamboost
from bamboost.common.file_handler import FileHandler, with_file_open

__all__ = [
    "BasePointer",
    "Group",
    "MutableGroup",
    "Dataset",
]

log = bamboost.BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])


class BasePointer:
    """Pointer to a location in an hdf5 file. The constructor takes a
    :class:`~.file_handler.FileHandler` and the in-file path to the object.
    The base class represents a generic group in the file

    Args:
        file_handler: file this belongs to
        path_to_data: infile path to object
    """

    @classmethod
    def new_pointer(cls, file_handler: FileHandler, path_to_data: str) -> BasePointer:
        """Returns a new pointer object."""
        with file_handler("r") as f:
            obj = f.file_object[path_to_data]
            if isinstance(obj, h5py.Group):
                if issubclass(cls, Group):
                    return cls(file_handler, path_to_data)
                else:
                    return Group(file_handler, path_to_data)
            elif isinstance(obj, h5py.Dataset):
                return Dataset(file_handler, path_to_data)
            else:
                return BasePointer(file_handler, path_to_data)

    def __init__(self, file_handler: FileHandler, path_to_data: str) -> None:
        self._file = file_handler
        self.path_to_data = path_to_data

    @property
    def obj(self):
        """The object this BasePointer points to. File needs to be open for access."""
        return self._file[self.path_to_data]

    @with_file_open("r")
    def __str__(self) -> str:
        try:
            return (
                f"{self.__class__} pointing to: "
                + self._file[self.path_to_data].__repr__()
            )
        except KeyError:
            return f"{self.__class__} pointing to: {self.path_to_data} (No data at that location!)"

    __repr__ = __str__

    @with_file_open("r")
    def __getattr__(self, __name: str) -> Any:
        """Any attribute request is sent to the h5py object the pointer points to."""
        if hasattr(self.obj, __name):
            return self.obj.__getattribute__(__name)
        return self.__getattribute__(__name)

    @with_file_open("r")
    def __getitem__(self, key):
        new_path = f"{self.path_to_data}/{key}"
        return self.new_pointer(self._file, new_path)

    @property
    @with_file_open()
    def attrs(self):
        return dict(self.obj.attrs)


class Group(BasePointer):
    def __init__(self, file_handler: FileHandler, path_to_data: str) -> None:
        super().__init__(file_handler, path_to_data)

    def _ipython_key_completions_(self):
        return self.keys()

    def __iter__(self):
        for key in self.keys():
            yield self.__getitem__(key)

    @with_file_open("r")
    def keys(self) -> set:
        return set(self.obj.keys())

    @with_file_open("r")
    def groups(self) -> set:
        return {key for key in self.keys() if isinstance(self.obj[key], h5py.Group)}

    @with_file_open("r")
    def datasets(self) -> set:
        return {key for key in self.keys() if isinstance(self.obj[key], h5py.Dataset)}

    @with_file_open("r")
    def extract_attrs(
        self, variant: Literal["group", "dataset", "all"] = "all"
    ) -> dict:
        """Extract the attributes of all members of the group.

        Args:
            variant: one of 'group', 'dataset', 'all'
        """
        if variant == "group":
            keys = self.groups()
        elif variant == "dataset":
            keys = self.datasets()
        elif variant == "all":
            keys = self.keys()
        else:
            raise ValueError("variant must be one of 'group', 'dataset', 'all'")

        attrs = {}
        for key in keys:
            attrs[key] = dict(self.obj[key].attrs)
        return attrs

    @with_file_open("r")
    def _repr_html_(self):
        """Repr showing the content of the group."""
        html_string = pkgutil.get_data(
            bamboost.__name__, "html/hdf5_group.html"
        ).decode()
        icon = pkgutil.get_data(bamboost.__name__, "html/icon.txt").decode()

        attrs_html = ""
        for key, val in self.attrs.items():
            attrs_html += f"""
            <tr>
                <td>{key}</td>
                <td>{val}</td>
            </tr>
            """
        groups_html = ""
        for key in self.groups():
            groups_html += f"""
            <tr>
                <td>{key}</td>
                <td>{len(self.obj[key])}</td>
            </tr>
            """
        ds_html = ""
        for key in self.datasets():
            obj = self.obj[key]
            ds_html += f"""
            <tr>
                <td>{key}</td>
                <td>{obj.dtype}</td>
                <td>{obj.shape}</td>
            </tr>
            """
        path = self.path_to_data
        if not path.startswith("/"):
            path = f"/{path}"
        return (
            html_string.replace("$NAME", path)
            .replace("$UID", self._file.simulation_uid)
            .replace("$ICON", icon)
            .replace("$attrs", attrs_html)
            .replace("$groups", groups_html)
            .replace("$datasets", ds_html)
            .replace("$version", bamboost.__version__)
        )


class MutableGroup(Group):
    """Used for the `userdata` group."""

    def __init__(self, file_handler: FileHandler, path_to_data: str) -> None:
        super().__init__(file_handler, path_to_data)
        # Create group if it doesn't exist
        with self._file("a", driver="mpio"):
            self._file.file_object.require_group(path_to_data)

    def _ipython_key_completions_(self):
        return self.keys().union(set(self.attrs.keys()))

    @with_file_open("r")
    def __getitem__(self, key) -> Any:
        """Used to access datasets (:class:`~bamboost.common.hdf_pointer.Dataset`)
        or groups inside this group (:class:`~bamboost.common.hdf_pointer.MutableGroup`)
        """
        try:
            return super().__getitem__(key)
        except KeyError:
            pass

        try:
            return self.obj.attrs[key]
        except KeyError:
            pass

        return super().__getitem__(key)

    def __setitem__(self, key, newvalue):
        """Used to set an attribute.
        Will be written as an attribute to the group.
        """
        if isinstance(newvalue, str) or not isinstance(newvalue, Iterable):
            self.update_attrs({key: newvalue})
        else:
            self.add_dataset(key, np.array(newvalue))

    @with_file_open("a")
    def __delitem__(self, key) -> None:
        """Deletes an item."""
        if key in self.attrs.keys():
            del self.obj.attrs[key]
        else:
            del self.obj[key]

    @with_file_open("a")
    def update_attrs(self, attrs: dict) -> None:
        """Update the attributes of the group.

        Args:
            attrs: the dictionary to write as attributes
        """
        self.obj.attrs.update(attrs)

    def add_dataset(
        self, name: str, vector: np.ndarray, attrs: dict = None, dtype: str = None
    ) -> None:
        """Add a dataset to the group. Error is thrown if attempting to overwrite
        with different shape than before. If same shape, data is overwritten
        (this is inherited from h5py -> require_dataset)

        Args:
            name: Name for the dataset
            vector: Data to write (max 2d)
            attrs: Optional. Attributes of dataset.
            dtype: Optional. dtype of dataset. If not specified, uses dtype of inpyt array
        """
        if attrs is None:
            attrs = {}
        length_local = vector.shape[0]
        length_p = np.array(self._file._comm.allgather(length_local))
        length = np.sum(length_p)
        dim = vector.shape[1:]
        vec_shape = length, *dim

        ranks = np.array([i for i in range(self._file._comm.size)])
        idx_start = np.sum(length_p[ranks < self._file._comm.rank])
        idx_end = idx_start + length_local

        with self._file("a", driver="mpio"):
            dataset = self.obj.require_dataset(
                name, shape=vec_shape, dtype=dtype if dtype else vector.dtype
            )
            dataset[idx_start:idx_end] = vector
            for key, item in attrs.items():
                dataset.attrs[key] = item

        log.info(f"Written {name} as userdata to {self._file.file_name}...")

    def require_group(self, name: str) -> MutableGroup:
        """Add a new group to the current group. If exists, return existing.

        Returns:
            :class:`~bamboost.hdf_pointer.Group`
        """
        return MutableGroup(self._file, f"{self.path_to_data}/{name}")


class Dataset(BasePointer):
    def __init__(self, file_handler: FileHandler, path_to_data: str) -> None:
        super().__init__(file_handler, path_to_data)

    @with_file_open("r")
    def __getitem__(self, slice):
        return self.obj[slice]

    @property
    @with_file_open()
    def attrs(self):
        return dict(self.obj.attrs)

    @property
    @with_file_open()
    def shape(self):
        return self.obj.shape

    @property
    @with_file_open()
    def dtype(self):
        return self.obj.dtype



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2024 Flavio Lorez and contributors
#
# There is no warranty for this code
import os
from typing import Any, Union

from bamboost import BAMBOOST_LOGGER
from bamboost._config import config
from bamboost.common._mock_mpi import MockMPI

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])

MPIType = Union[MockMPI, Any]

MPI_ON = config.get("options", {}).get("mpi", True)
ENV_BAMBOOST_MPI: bool = os.environ.get("BAMBOOST_MPI", None)
"""Indicates the use of `mpi4py.MPI`. If `0`, the `MockMPI` class is used
instead. Is set by reading the environment variable `BAMBOOST_MPI` [0 or 1].
"""
if ENV_BAMBOOST_MPI is not None:
    MPI_ON = ENV_BAMBOOST_MPI == "1"


def _get_mpi_module():
    if not MPI_ON:
        return MockMPI

    try:
        from mpi4py import MPI

        return MPI
    except ImportError:
        log.info("`mpi4py` unavailable [using a mock MPI module]")
        return MockMPI


MPI: Union[MockMPI, Any] = _get_mpi_module()



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

from contextlib import contextmanager
from enum import Enum
from typing import Literal, TypedDict

import numpy as np

from bamboost.common.file_handler import open_h5file
from bamboost.common.mpi import MPI
from bamboost.simulation_writer import SimulationWriter

try:
    import fenics as fe
except ImportError:
    raise ImportError("FEniCS not found. Module unavailable.")

__all__ = ["FenicsWriter"]


class WriteStrategy(Enum):
    """Enum for write style."""

    SCATTERED = 0
    CONTIGUOUS = 1


class FenicsWriter(SimulationWriter):
    """
    Helper writer for input from FEniCS directly.

    Args:
        uid: Unique identifier for the simulation
        path: Path to database
        comm: MPI communicator
        create_if_not_exists: Create file if it does not exist
        write_strategy: Write strategy for the data. Contiguous is faster but
            requires the entire array to fit in memory of the root process.
    """

    def __init__(
        self,
        uid: str,
        path: str,
        comm: MPI.Comm = MPI.COMM_WORLD,
        create_if_not_exists: bool = False,
        write_strategy: WriteStrategy = WriteStrategy.SCATTERED,
    ):
        super().__init__(uid, path, comm)
        self.write_strategy = WriteStrategy(write_strategy)

    def add_field(
        self,
        name: str,
        func: fe.Function,
        time: float = None,
        mesh: str = None,
        dtype: str = None,
        center: Literal["Node", "Cell"] = "Node",
    ) -> None:
        """Add a dataset to the file. The data is stored at `data/`.

        Args:
            name: Name for the dataset
            func: FEniCS function to store
            time: Optional. time
            mesh: Optional. Linked mesh for this data
            dtype: Optional. Numpy style datatype, see h5py documentation,
                defaults to the dtype of the vector
            center: Optional. Center of the data. Can be 'Node' or 'Cell'.
                Default is 'Node'.
        """
        mesh = mesh if mesh is not None else self._default_mesh
        time = time if time is not None else self.step

        {
            WriteStrategy.SCATTERED: self._dump_fenics_field,
            WriteStrategy.CONTIGUOUS: self._dump_fenics_field_on_root,
        }.get(self.write_strategy)(
            f"data/{name}/{self.step}",
            func,
            dtype=dtype,
            center=center,
        )
        self._comm.barrier()  # attempt to fix bug (see SimulationWriter add_field)

        if self._prank == 0:
            with self._file("a"):
                vec = self._file["data"][name][str(self.step)]
                vec.attrs.update({"center": center, "mesh": mesh, "t": time})

        self._comm.barrier()  # attempt to fix bug (see SimulationWriter add_field)

    def _dump_fenics_field(
        self,
        location: str,
        field: fe.Function,
        dtype: str = None,
        center: Literal["Node", "Cell"] = "Node",
    ) -> None:
        # get global dofs ordering and vector
        if center == "Node":
            data = self._get_global_dofs(field)
        elif center == "Cell":
            data = self._get_global_dofs_cell_data(field)
        else:
            raise ValueError("Center must be 'Node' or 'Cell'.")

        vector = data["vector"]
        global_map = data["global_map"]
        global_size = data["global_size"]

        dim = data["vector"].shape[1:] if data["vector"].ndim > 1 else None

        group_name, dataset_name = location.rstrip("/").rsplit("/", 1)

        # Write vector to file
        with self._file("a", driver="mpio", comm=self._comm) as f:
            f = f.file_object
            grp = f.require_group(group_name)
            vec = grp.require_dataset(
                dataset_name,
                shape=(global_size, *dim) if dim else (global_size,),
                dtype=dtype if dtype else vector.dtype,
            )
            vec[global_map] = vector

    def _dump_fenics_field_on_root(
        self,
        location: str,
        field: fe.Function,
        dtype: str = None,
        center: Literal["Node", "Cell"] = "Node",
    ) -> None:
        """Assembles the vector on the root process and writes it to file contiguously.

        This is faster but requires the entire array to fit in memory.

        Args:
            location: Location in the HDF5 file
            field: FEniCS function
            dtype: Optional. Numpy style datatype, see h5py documentation,
                defaults to the dtype of the vector
            center: Optional. Center of the data. Can be 'Node' or 'Cell'.
                Default is 'Node'.
        """
        # get global dofs ordering and vector
        if center == "Node":
            data = self._get_global_dofs(field)
        elif center == "Cell":
            data = self._get_global_dofs_cell_data(field)
        else:
            raise ValueError("Center must be 'Node' or 'Cell'.")

        vector = data["vector"]
        global_map = data["global_map"]
        global_size = data["global_size"]

        dim = data["vector"].shape[1:] if data["vector"].ndim > 1 else None

        group_name, dataset_name = location.rstrip("/").rsplit("/", 1)

        vector_p = self._comm.gather(vector)
        global_map_p = self._comm.gather(global_map)

        # On RAM, construct a contiguous vector on the root process
        if self._prank == 0:
            vector_contiguous = np.zeros(
                (global_size, *dim) if dim else (global_size,), dtype=vector.dtype
            )
            for map, vec in zip(global_map_p, vector_p):
                vector_contiguous[map] = vec

            # Write vector to file
            with self._file("a") as f:
                grp = f.require_group(group_name)
                vec = grp.require_dataset(
                    dataset_name,
                    shape=(global_size, *dim) if dim else (global_size,),
                    dtype=dtype if dtype else vector.dtype,
                )
                vec[:] = vector_contiguous

    class FenicsFieldInformation(TypedDict):
        vector: np.ndarray
        global_map: np.ndarray
        global_size: int

    def _get_global_dofs(self, func: fe.Function) -> FenicsFieldInformation:
        """
        Get global dofs for a given function.

        Args:
            - func: Expression/field/function

        Returns:
            A dict with the local vector, a mapping from the local to the
            global vertices (both sorted by index because h5py complains if
            slicing is not continuously increasing), and the number of global
            vertices.
        """
        assert hasattr(
            func, "function_space"
        ), "Input is likely an indexed coefficient. Project to it's own function space first."

        # Project to CG1 if necessary
        if func.ufl_element().degree() != 1:
            func = fe.project(
                func, fe.FunctionSpace(func.function_space().mesh(), "CG", 1)
            )

        mesh = func.function_space().mesh()
        global_size = mesh.num_entities_global(0)
        shape = (-1, *func.ufl_shape) if func.ufl_shape else (-1,)
        val_per_vertex = np.prod(shape[1:]).astype(np.int32) if func.ufl_shape else 1

        dofmap = func.function_space().dofmap()
        d2v = fe.dof_to_vertex_map(func.function_space())
        d2v = (
            d2v[np.arange(0, len(d2v), val_per_vertex, dtype=np.int32)]
            // val_per_vertex
        )

        loc0, loc1 = (i // val_per_vertex for i in dofmap.ownership_range())
        global_vertex_numbers = mesh.topology().global_indices(0)
        global_vertices = global_vertex_numbers[d2v[: loc1 - loc0]]
        sort_indices = np.argsort(global_vertices)

        local_vector = (
            func.vector().get_local().reshape(shape)[: loc1 - loc0][sort_indices]
        )

        return {
            "vector": local_vector,
            "global_map": global_vertices[sort_indices],
            "global_size": global_size,
        }

    def _get_global_dofs_cell_data(self, func: fe.Function) -> FenicsFieldInformation:
        """
        Get global dofs for a given function.

        Args:
            - func: Expression/field/function

        Returns:
            A dict with the local vector, a mapping from the local to the
            global vertices (both sorted by index because h5py complains if
            slicing is not continuously increasing), and the number of global
            vertices.
        """
        V = func.function_space()
        val_per_vertex = (
            np.prod(func.ufl_shape).astype(np.int32) if func.ufl_shape else 1
        )

        mesh = V.mesh()
        dofmap = V.dofmap()

        local_to_global_indices = dofmap.tabulate_local_to_global_dofs()
        local_to_global_indices = (
            local_to_global_indices[
                np.arange(
                    0, len(local_to_global_indices), val_per_vertex, dtype=np.int32
                )
            ]
            // val_per_vertex
        )
        shape = (-1, *func.ufl_shape) if func.ufl_shape else (-1,)
        local_vector = func.vector().get_local().reshape(shape)

        return {
            "vector": local_vector,
            "global_map": local_to_global_indices,
            "global_size": self._comm.allreduce(mesh.num_cells(), op=MPI.SUM),
        }

    def add_mesh(self, mesh: fe.Mesh, mesh_name: str = None) -> None:
        """
        Add the mesh to file using fe.HDF5File. I can't figure out how to
        extract the local mesh data in correct order when running in parallel.

        Args:
            mesh: FEniCS mesh object
            mesh_name: name for mesh (default = `mesh`)
        """
        mesh_name = mesh_name if mesh_name is not None else self._default_mesh
        mesh_location = f"{self._mesh_location}/{mesh_name}/"

        assert not self._file.file_object, "File is open -> Quitting"

        with self.temporary_close_file():
            with fe.HDF5File(self._comm, self.h5file, "a") as f:
                f.write(mesh, mesh_location)

    @contextmanager
    def temporary_close_file(self):
        was_open = False
        if self._file.file_object:
            self._file.file_object.close()
            was_open = True
        try:
            yield
        finally:
            if was_open:
                self._file.file_object = open_h5file(
                    self._file.file_name,
                    self._file.mode,
                    self._file.driver,
                    self._file.comm,
                )



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code
from __future__ import annotations

import numbers
import os
import pkgutil
import shutil
import uuid
from ctypes import ArgumentError
from typing import Any, Generator, Iterable, Union

import h5py
import numpy as np
import pandas as pd

from bamboost import BAMBOOST_LOGGER, index
from bamboost.common.file_handler import open_h5file
from bamboost.common.mpi import MPI
from bamboost.common.utilities import flatten_dict
from bamboost.index import DatabaseTable, IndexAPI, config
from bamboost.simulation import Simulation
from bamboost.simulation_writer import SimulationWriter

__all__ = [
    "Manager",
    "ManagerFromUID",
    "ManagerFromName",
]

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])


class ManagerFromUID(object):
    """Get a database by its UID. This is used for autocompletion in ipython."""

    def __init__(self) -> None:
        # or [] to circumvent Null type (MPI)
        ids = IndexAPI().fetch(f"SELECT id, path FROM dbindex") or []
        self.completion_keys = tuple(
            [f'{key} - {"..."+val[-25:] if len(val)>=25 else val}' for key, val in ids]
        )

    def _ipython_key_completions_(self):
        return self.completion_keys

    def __getitem__(self, key) -> Manager:
        key = key.split()[0]  # take only uid
        return Manager(uid=key, create_if_not_exist=False)


class ManagerFromName(object):
    """Get a database by its path/name. This is used for autocompletion in ipython."""

    def __init__(self) -> None:
        paths = IndexAPI().fetch("SELECT path FROM dbindex") or []
        self.completion_keys = tuple(paths)

    def _ipython_key_completions_(self):
        return self.completion_keys

    def __getitem__(self, key) -> Manager:
        return Manager(key, create_if_not_exist=False)


class Manager:
    """View of database.

    Args:
        path (`str`): path to the directory of the database. If doesn't exist,
            a new database will be created.
        comm (`MPI.Comm`): MPI communicator
        uid: UID of the database

    Attributes:
        FIX_DF: If False, the dataframe of the database is reconstructed every
            time it is accessed.
        fromUID: Access a database by its UID
        fromName: Access a database by its path/name

    Example:
        >>> db = Manager("path/to/db")
        >>> db.df # DataFrame of the database
    """

    FIX_DF = True
    fromUID: ManagerFromUID = ManagerFromUID()
    fromName: ManagerFromName = ManagerFromName()

    def __init__(
        self,
        path: str = None,
        comm: MPI.Comm = MPI.COMM_WORLD,
        uid: str = None,
        create_if_not_exist: bool = True,
    ):
        # provided uid has precedence
        if uid is not None:
            path = self._index.get_path(uid.upper())
            path = comm.bcast(path, root=0)
        self.path = path
        self.comm = comm

        # check if path exists
        if not os.path.isdir(path):
            if not create_if_not_exist:
                raise NotADirectoryError("Specified path is not a valid path.")
            log.info(f"Created new database ({path})")
            self._make_new(path)

        # retrieve the UID of the database from the id file
        # if not found, a new one is generated
        self.UID = uid or self._retrieve_uid()

        # Update the SQL table for the database
        try:
            with self._index.open():
                self._index.insert_path(self.UID, self.path)
                self._table.create_database_table()
                self._table.sync()
        except index.Error as e:
            log.warning(f"index error: {e}")

    def __getitem__(self, key: Union[str, int]) -> Simulation:
        """Returns the simulation in the specified row of the dataframe.

        Args:
            key: The simulation identifier (`str`) or the row index (`int`).
        Returns:
            The selected simulation object.
        """
        if isinstance(key, str):
            return self.sim(key)
        else:
            return self.sim(self.df.loc[key, "id"])

    def _repr_html_(self) -> str:
        """HTML repr for ipython/notebooks. Uses string replacement to fill the
        template code.
        """
        html_string = pkgutil.get_data(__name__, "html/manager.html").decode()
        icon = pkgutil.get_data(__name__, "html/icon.txt").decode()
        return (
            html_string.replace("$ICON", icon)
            .replace("$db_path", self.path)
            .replace("$db_uid", self.UID)
            .replace("$db_size", str(len(self)))
        )

    def __len__(self) -> int:
        return len(self.all_uids)

    def __iter__(self) -> Generator[Simulation, None, None]:
        for sim in self.sims():
            yield sim

    def _ipython_key_completions_(self):
        return self.all_uids

    @property
    def _index(self) -> IndexAPI:
        """The index which contains this database."""
        return IndexAPI()

    @property
    def _table(self) -> DatabaseTable:
        """The table in the sql database for this database."""
        return self._index.get_database_table(self.UID)

    def _retrieve_uid(self) -> str:
        """Get the UID of this database from the file tree."""
        try:
            return index.get_uid_from_path(self.path)
        except FileNotFoundError:
            pass

        log.warning("Database exists but no UID found. Generating new UID.")
        return self._make_new(self.path)

    def _make_new(self, path) -> str:
        """Initialize a new database."""
        from datetime import datetime

        # Create directory for database
        os.makedirs(path, exist_ok=True)

        # Assign a unique id to the database
        self.UID = f"{uuid.uuid4().hex[:10]}".upper()
        uid_file = os.path.join(path, f".BAMBOOST-{self.UID}")
        with open(uid_file, "a") as f:
            f.write(self.UID + "\n")
            f.write(f'Date of creation: {datetime.now().strftime("%d/%m/%Y %H:%M:%S")}')
        os.chmod(uid_file, 0o444)  # read only for uid file

        log.info(f"Registered new database (uid = {self.UID})")
        self._index.insert_path(self.UID, path)
        return self.UID

    @property
    def all_uids(self) -> list:
        if not self.FIX_DF or not hasattr(self, "_all_uids"):
            self._all_uids = self._get_uids()
        return self._all_uids

    @all_uids.setter
    def all_uids(self, value: set | list):
        self._all_uids = value

    def _get_uids(self) -> list:
        """Get all simulation names in the database."""
        all_uids = list()
        for dir in os.listdir(self.path):
            if not os.path.isdir(os.path.join(self.path, dir)):
                continue
            if any(
                [i.endswith(".h5") for i in os.listdir(os.path.join(self.path, dir))]
            ):
                all_uids.append(dir)
        return all_uids

    @property
    def df(self) -> pd.DataFrame:
        """View of the database and its parametric space.

        Returns:
            :class:`pd.DataFrame`
        """
        if not hasattr(self, "_dataframe"):
            return self.get_view()
        if self.FIX_DF and self._dataframe is not None:
            return self._dataframe
        return self.get_view()

    def _get_parameters_for_uid(
        self, uid: str, include_linked_sims: bool = False
    ) -> dict:
        """Get the parameters for a given uid.

        Args:
            uid (`str`): uid of the simulation
            include_linked_sims (`bool`): if True, include the parameters of linked sims
        """
        h5file_for_uid = os.path.join(self.path, uid, f"{uid}.h5")
        tmp_dict = dict()

        with open_h5file(h5file_for_uid, "r") as f:
            if "parameters" in f.keys():
                tmp_dict.update(f["parameters"].attrs)
            if "additionals" in f.keys():
                tmp_dict.update({"additionals": dict(f["additionals"].attrs)})
            tmp_dict.update(f.attrs)

        if include_linked_sims:
            for linked, full_uid in self.sim(uid).links.attrs.items():
                sim = Simulation.fromUID(full_uid)
                tmp_dict.update(
                    {f"{linked}.{key}": val for key, val in sim.parameters.items()}
                )
        return tmp_dict

    def get_view(self, include_linked_sims: bool = False) -> pd.DataFrame:
        """View of the database and its parametric space. Read from the sql
        database. If `include_linked_sims` is True, the individual h5 files are
        scanned.

        Args:
            include_linked_sims: if True, include the parameters of linked sims

        Examples:
            >>> db.get_view()
            >>> db.get_view(include_linked_sims=True)
        """
        if include_linked_sims:
            return self.get_view_from_hdf_files(include_linked_sims=include_linked_sims)

        try:
            with self._table.open():
                self._table.sync()
                df = self._table.read_table()
        except index.Error as e:
            log.warning(f"index error: {e}")
            return self.get_view_from_hdf_files(include_linked_sims=include_linked_sims)

        if df.empty:
            return df
        df["time_stamp"] = pd.to_datetime(df["time_stamp"])

        # Sort dataframe columns
        columns_start = ["id", "notes", "status", "time_stamp"]
        columns_start = [col for col in columns_start if col in df.columns]
        self._dataframe = df[[*columns_start, *df.columns.difference(columns_start)]]

        opts = config.get("options", {})
        if "sort_table_key" in opts:
            self._dataframe.sort_values(
                opts.get("sort_table_key", "id"),
                ascending=opts.get("sort_table_order", "asc") == "asc",
                inplace=True,
            )
        return self._dataframe

    def get_view_from_hdf_files(
        self, include_linked_sims: bool = False
    ) -> pd.DataFrame:
        """View of the database and its parametric space. Read from the h5
        files metadata.

        Args:
            include_linked_sims: if True, include the parameters of linked sims
        """
        all_uids = self._get_uids()
        data = list()

        for uid in all_uids:
            tmp_dict = self._get_parameters_for_uid(
                uid, include_linked_sims=include_linked_sims
            )
            data.append(tmp_dict)

        df = pd.DataFrame.from_records(data)
        if df.empty:
            return df
        df["time_stamp"] = pd.to_datetime(df["time_stamp"])

        # Sort dataframe columns
        columns_start = ["id", "notes", "status", "time_stamp"]
        columns_start = [col for col in columns_start if col in df.columns]
        self._dataframe = df[[*columns_start, *df.columns.difference(columns_start)]]
        return self._dataframe

    @property
    def data_info(self) -> pd.DataFrame:
        """Return view of stored data for all simulations

        Returns:
            :class:`pd.DataFrame`
        """
        data = list()
        for uid in self.all_uids:
            h5file_for_uid = os.path.join(self.path, uid, f"{uid}.h5")
            with open_h5file(h5file_for_uid, "r") as file:
                try:
                    tmp_dict = dict()
                    tmp_dict = {
                        key: (
                            len(file[f"data/{key}"]),
                            file[f"data/{key}/0"].shape,
                            file[f"data/{key}/0"].dtype,
                        )
                        for key in file["data"].keys()
                    }
                except KeyError:
                    tmp_dict = dict()
                data.append(tmp_dict)
        return pd.DataFrame.from_records(data)

    def sim(
        self,
        uid: str,
        return_writer: bool = False,
        writer_type: SimulationWriter = SimulationWriter,
    ) -> Simulation:
        """Get an existing simulation with uid. Same as accessing with `db[uid]` directly.

        Args:
            uid (`str`): unique identifier
            return_writer: if true, return `SimulationWriter`, otherwise
                return `Simulation`
            writer_type: Optionally, you can specify a custom writer type to return.

        Returns:
            :class:`~bamboost.simulation.Simulation`
        """
        if return_writer:
            return writer_type(uid, self.path, self.comm)
        return Simulation(uid, self.path, self.comm, _db_id=self.UID)

    def sims(
        self,
        select: pd.Series | pd.DataFrame | dict = None,
        sort: str = None,
        reverse: bool = False,
        exclude: set = None,
        return_writer: bool = False,
    ) -> list[Simulation]:
        """Get all simulations in a list. Optionally, get all simulations matching the
        given selection using pandas.

        Args:
            select: Selection of simulations. Can be one of the following.
                - Pandas boolean series: A boolean series with the same length as the dataframe.
                - Pandas DataFrame: A subset of the full dataframe.
                - Dictionary: A dictionary with the parameters to select (see `find` for details).
            sort (`str`): Optionally sort the list with this keyword
            reverse (`bool`): swap sort direction
            exclude (`list[str]`): sims to exclude
            return_writer: if true, return `SimulationWriter`, otherwise
                return `Simulation`

        Returns:
            A list of `:class:~bamboost.simulation.Simulation` objects

        Examples:
            >>> db.sims(select=db.df["status"] == "finished", sort="time_stamp")
        """
        if select is None:
            id_list = self.all_uids
        elif isinstance(select, pd.DataFrame):
            id_list = select["id"].values
        elif isinstance(select, pd.Series):
            id_list = self.df[select]["id"].values
        elif isinstance(select, dict):
            id_list = self.find(select)["id"].values
        else:
            raise ArgumentError('Invalid argument for argument "select"')

        if exclude is not None:
            exclude = list([exclude]) if isinstance(exclude, str) else exclude
            id_list = [id for id in id_list if id not in exclude]

        existing_sims = [self.sim(uid, return_writer) for uid in id_list]

        if sort is None:
            return existing_sims
        else:
            return sorted(
                existing_sims, key=lambda s: s.parameters[sort], reverse=reverse
            )

    def create_simulation(
        self,
        uid: str = None,
        parameters: dict = None,
        skip_duplicate_check: bool = False,
        *,
        prefix: str = None,
        duplicate_action: str = "prompt",
        note: str = None,
        files: list[str] = None,
        links: dict[str, str] = None,
    ) -> SimulationWriter:
        """Get a writer object for a new simulation. This is written for paralell use
        as it is likely that this may be used in an executable, creating multiple runs
        for a parametric space, which may be run in paralell.

        Args:
            uid (`str`): The name/uid for the simulation. If not specified, a random id
                will be assigned.
            parameters (`dict`): Parameter dictionary. If provided, the parameters will be
                checked against the existing sims for duplication. Otherwise, they may be
                specified later with `bamboost.simulation_writer.SimulationWriter.add_parameters`.
                Note:
                    The parameters are stored in the h5 file as attributes.
                    - If the value is a dict, it is flattened using
                      `bamboost.common.utilities.flatten_dict`.
                    - If the value is a list/array, it is stored as a dataset.
            skip_duplicate_check (`bool`): if True, the duplicate check is skipped.
            prefix (`str`): Prefix for the uid. If not specified, no prefix is used.
            duplicate_action (`str`): how to deal with duplicates. Replace
                first duplicate ('r'), Create with altered uid (`c`), Create new
                with new id (`n`), Abort (`a`) default "prompt" for each
                duplicate on a case by case basis.
            note (`str`): Note for the simulation.
            files (`list`): List of files to copy to the simulation directory.
            links (`dict`): Dictionary of links to other simulations.

        Note:
            The files and links are copied to the simulation directory. The files are
            copied with the same name as the original file. The links are copied with
            the given name.

        Examples:
            >>> db.create_simulation(parameters={"a": 1, "b": 2})

            >>> db.create_simulation(uid="my_sim", parameters={"a": 1, "b": 2}, prefix="test")

        Returns:
            A simulation writer object.
        """
        if parameters and not skip_duplicate_check:
            go_on = True
            if self.comm.rank == 0:
                go_on, uid = self._check_duplicate(
                    parameters, uid, duplicate_action=duplicate_action
                )
            self.comm.bcast((go_on, uid), root=0)
            if not go_on:
                print("Aborting by user desire...")
                return None

        if self.comm.rank == 0:
            if not uid:
                uid = uuid.uuid4().hex[:8]  # Assign random unique identifier
            if isinstance(prefix, str) and prefix != "":
                uid = "_".join([prefix, uid])
        uid = self.comm.bcast(uid, root=0)

        try:
            # Create directory and h5 file
            if self.comm.rank == 0:
                os.makedirs(os.path.join(self.path, uid), exist_ok=True)
                path_to_h5_file = os.path.join(self.path, uid, f"{uid}.h5")
                if os.path.exists(path_to_h5_file):
                    os.remove(path_to_h5_file)
                h5py.File(path_to_h5_file, "a").close()  # create file

            new_sim = SimulationWriter(uid, self.path, self.comm)
            new_sim.initialize()  # sets metadata and status
            # add the id to the (fixed) _all_uids list
            if hasattr(self, "_all_uids"):
                self._all_uids.append(new_sim.uid)

            # Add parameters, note, files, and links
            if not any([parameters, note, files, links]):
                return new_sim

            with new_sim._file("r+"):
                if parameters:
                    new_sim.add_parameters(parameters)
                if note:
                    new_sim.change_note(note)
                if files:
                    new_sim.copy_file(files)
                if links:
                    [
                        new_sim.links.__setitem__(name, uid)
                        for name, uid in links.items()
                    ]

            return new_sim

        except Exception as e:
            # If any error occurs, remove the partially created simulation
            if self.comm.rank == 0:
                self.remove(uid)
            raise e  # Re-raise the exception after cleanup

    def remove(self, uid: str) -> None:
        """CAUTION, DELETING DATA. Remove the data of a simulation.

        Args:
            uid (`str`): uid
        """
        shutil.rmtree(os.path.join(self.path, uid))
        self._table.sync()

    def find(self, parameter_selection: dict[str, Any]) -> pd.DataFrame:
        """Find simulations with the given parameters.

        The dictionary can contain callables to filter inequalities or other
        filters.

        Examples:
            >>> db.find({"a": 1, "b": lambda x: x > 2})
            >>> db.find({"a": 1, "b": 2})

        Args:
            parameter_selection (dict): parameter selection dictionary
        """
        parameter_selection = flatten_dict(parameter_selection)
        params = {}
        filters = {}
        for key, val in parameter_selection.items():
            if callable(val):
                filters[key] = val
            else:
                params[key] = val

        df = self.df
        matches = self._list_duplicates(params, df=df)
        matches = df[df.id.isin(matches)]
        if len(matches) == 0:
            return matches

        for key, func in filters.items():
            matches = matches[matches[key].apply(func)]

        return matches

    def _list_duplicates(
        self, parameters: dict, *, df: pd.DataFrame = None
    ) -> list[str]:
        """List ids of duplicates of the given parameters.

        Args:
            parameters (dict): parameter dictionary
            df (pd.DataFrame): dataframe to search in. If not provided, the
                dataframe from the sql database is used.
        """
        if df is None:
            df: pd.DataFrame = self._table.read_table()
        params = flatten_dict(parameters)

        class ComparableIterable:
            def __init__(self, ori):
                self.ori = np.asarray(ori)

            def __eq__(self, other):
                other = np.asarray(other)
                if other.shape != self.ori.shape:
                    return False
                return (other == self.ori).all()

        # make all iterables comparable by converting them to ComparableIterable
        for k in params.keys():
            if isinstance(params[k], Iterable) and not isinstance(params[k], str):
                params[k] = ComparableIterable(params[k])

        # if any of the parameters is not in the dataframe, no duplicates
        for p in params:
            if p not in df.keys():
                return []

        # get matching rows where all values of the series are equal to the corresponding values in the dataframe
        s = pd.Series(params)
        match = df.loc[(df[s.keys()].apply(lambda row: (s == row).all(), axis=1))]
        return match.id.tolist()

    def _check_duplicate(
        self, parameters: dict, uid: str, duplicate_action: str = "prompt"
    ) -> tuple:
        """Checking whether the parameters dictionary exists already.
        May need to be improved...

        Args:
            parameters (`dict`): parameter dictionary to check for
            uid (`str`): uid
        Returns:
            Tuple(Bool, uid) wheter to continue and with what uid.
        """

        duplicates = self._list_duplicates(parameters)

        if not duplicates:
            return True, uid

        print(
            "The parameter space already exists. Here are the duplicates:",
            flush=True,
        )
        print(self.df[self.df["id"].isin([i for i in duplicates])], flush=True)

        if duplicate_action == "prompt":
            # What should be done?
            prompt = input(
                "Replace first duplicate ('r'), Create with altered uid (`c`), "
                + "Create new with new id (`n`), Abort (`a`): "
            )
        else:
            prompt = duplicate_action

        if prompt == "r":
            self.remove(duplicates[0])
            return True, uid
        if prompt == "a":
            return False, uid
        if prompt == "n":
            return True, uid
        if prompt == "c":
            return True, self._generate_subuid(duplicates[0].split(".")[0])

        raise ArgumentError("Answer not valid! Aborting")

    def _generate_subuid(self, uid_base: str) -> str:
        """Return a new sub uid for the base uid.
        Following the following format: `base_uid.1`

        Args:
            uid_base (`str`): base uid for which to find the next subid.
        Returns:
            New uid string
        """
        uid_list = [uid for uid in self.all_uids if uid.startswith(uid_base)]
        subiterator = max(
            [int(id.split(".")[1]) for id in uid_list if len(id.split(".")) > 1] + [0]
        )
        return f"{uid_base}.{subiterator+1}"

    def global_fields_in_all(self) -> list:
        """Get a list of all global fields in all simulations.

        Returns:
            List of global fields
        """
        fields = set()
        for sim in self:
            try:
                fields.update(sim.globals.columns)
            except KeyError:
                continue

        return fields

    def get_parameters(self) -> dict:
        """Get the parameters used in this database.

        Returns:
            Dictionary of parameters with it's count, range, and type. Sorted by count.
        """
        parameters = dict()
        for sim in self:
            for key, val in sim.parameters.items():
                if key not in parameters:
                    range = (val, val) if isinstance(val, numbers.Number) else None
                    parameters[key] = {"range": range, "count": 1, "type": type(val)}
                else:
                    if isinstance(val, numbers.Number):
                        parameters[key]["range"] = (
                            min(parameters[key]["range"][0], val),
                            max(parameters[key]["range"][1], val),
                        )
                    parameters[key]["count"] += 1
                    parameters[key]["type"] = type(val)
        return dict(
            sorted(parameters.items(), key=lambda x: x[1]["count"], reverse=True)
        )



class ManagerCachedFromUID(object):
    """Get a database by its UID. This is used for autocompletion in ipython."""

    def __init__(self) -> None:
        # or [] to circumvent Null type (MPI)
        ids = IndexAPI().fetch(f"SELECT id, path FROM dbindex") or []
        self.completion_keys = tuple(
            [f'{key} - {"..."+val[-25:] if len(val)>=25 else val}' for key, val in ids]
        )

    def _ipython_key_completions_(self):
        return self.completion_keys

    def __getitem__(self, key) -> ManagerCached:
        key = key.split()[0]  # take only uid
        return ManagerCached(uid=key)


class ManagerCachedFromName(object):
    """Get a database by its path/name. This is used for autocompletion in ipython."""

    def __init__(self) -> None:
        paths = IndexAPI().fetch("SELECT path FROM dbindex") or []
        self.completion_keys = tuple(paths)

    def _ipython_key_completions_(self):
        return self.completion_keys

    def __getitem__(self, key) -> ManagerCached:
        return ManagerCached(key)


class ManagerCached(Manager):
    """View of database that retrieves data from sql databases without updating it.

    Args:
        path (`str`): path to the directory of the database. If doesn't exist,
            a new database will be created.
        comm (`MPI.Comm`): MPI communicator
        uid: UID of the database

    Attributes:
        FIX_DF: If False, the dataframe of the database is reconstructed every
            time it is accessed.
        fromUID: Access a database by its UID
        fromName: Access a database by its path/name

    Example:
        >>> db = Manager("path/to/db")
        >>> db.df # DataFrame of the database
    """

    FIX_DF = True
    fromUID: ManagerCachedFromUID = ManagerCachedFromUID()
    fromName: ManagerCachedFromName = ManagerCachedFromName()

    def __init__(
        self,
        path: str = None,
        comm: MPI.Comm = MPI.COMM_WORLD,
        uid: str = None,
    ):
        # provided uid has precedence
        if uid is not None:
            path = self._index.get_path(uid.upper())
            path = comm.bcast(path, root=0)
        self.path = path
        self.comm = comm

        # check if path exists
        if not os.path.isdir(path):
            raise ValueError(f"path {path} does not exist")

        # retrieve the UID of the database from the id file
        # if not found, a new one is generated
        self.UID = uid or self._retrieve_uid()


    def get_view(self, include_linked_sims: bool = False) -> pd.DataFrame:
        """View of the database and its parametric space. Read from the sql
        database. If `include_linked_sims` is True, the individual h5 files are
        scanned.

        Args:
            include_linked_sims: if True, include the parameters of linked sims

        Examples:
            >>> db.get_view()
            >>> db.get_view(include_linked_sims=True)
        """
        if include_linked_sims:
            return self.get_view_from_hdf_files(include_linked_sims=include_linked_sims)

        try:
            with self._table.open():
                df = self._table.read_table()
        except index.Error as e:
            log.warning(f"index error: {e}, trying again in 1s")
            import time
            time.sleep(1)
            return self.get_view()

        if df.empty:
            return df
        df["time_stamp"] = pd.to_datetime(df["time_stamp"])

        # Sort dataframe columns
        columns_start = ["id", "notes", "status", "time_stamp"]
        columns_start = [col for col in columns_start if col in df.columns]
        self._dataframe = df[[*columns_start, *df.columns.difference(columns_start)]]

        opts = config.get("options", {})
        if "sort_table_key" in opts:
            self._dataframe.sort_values(
                opts.get("sort_table_key", "id"),
                ascending=opts.get("sort_table_order", "asc") == "asc",
                inplace=True,
            )
        return self._dataframe
    
    def _get_uids(self) -> list:
        """Get all simulation names in the database."""
        return self.df['id'].tolist()


import os
from unittest.mock import MagicMock

import numpy as np
import pytest
from test_manager import temp_manager


@pytest.mark.parametrize(
    "dtype,array",
    [
        ["<f4", np.array([1, 2], dtype=int)],
        ["int32", np.array([1, 2])],
    ],
)
def test_field_dtype_forced(temp_manager, dtype, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_field("test", array, dtype=dtype)

    sim.finish_step()
    sim.add_field("test", 2 * array, dtype=dtype)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    assert sim.data["test"].dtype == dtype


@pytest.mark.parametrize(
    "dtype,value",
    [
        ["<f4", 1],
        ["int32", 1],
    ],
)
def test_global_field_dtype_forced(temp_manager, dtype, value):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_global_field("test", value, dtype=dtype)

    sim.finish_step()
    sim.add_global_field("test", 2 * value, dtype=dtype)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    assert sim.globals["test"].dtype == dtype


@pytest.mark.parametrize(
    "array",
    [
        np.array([1.0, 2.0]),
        np.array([1, 2], dtype="int"),
    ],
)
def test_field_dtype_inferred(temp_manager, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_field("test", array)

    sim.finish_step()
    sim.add_field("test", 2 * array)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    # The dtypes are equivalent (although they may have another string representation)
    assert np.dtype(sim.data["test"].dtype) == array.dtype


@pytest.mark.parametrize(
    "val",
    [
        1.0,
        1,
    ],
)
def test_global_field_dtype_inferred(temp_manager, val):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_global_field("test", val)

    sim.finish_step()
    sim.add_global_field("test", 2 * val)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    # The dtypes are equivalent (although they may have another string representation)
    assert np.dtype(sim.globals["test"].dtype) == np.array([val]).dtype


@pytest.mark.parametrize(
    "array",
    [
        np.array([1.0, 2.0]),
        np.array([1, 2], dtype="int"),
    ],
)
def test_userdata_dtype_inferred(temp_manager, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.userdata.add_dataset("test", array)

    sim.finish_step()
    sim.userdata.add_dataset("test", 2 * array)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    # The dtypes are equivalent (although they may have another string representation)
    assert np.dtype(sim.userdata["test"].dtype) == array.dtype


@pytest.mark.parametrize(
    "dtype,array",
    [
        ["<f4", np.array([1, 2], dtype=int)],
        ["int32", np.array([1, 2])],
    ],
)
def test_userdata_dtype_forced(temp_manager, dtype, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.userdata.add_dataset("test", array, dtype=dtype)

    sim.finish_step()
    sim.userdata.add_dataset("test", 2 * array, dtype=dtype)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    assert sim.userdata["test"].dtype == dtype


def test_enter_path(temp_manager):
    sim = temp_manager.create_simulation()
    sim.finish_sim()
    path = sim.path
    uid = sim.uid
    original_path = os.getcwd()
    with sim.enter_path():
        assert os.getcwd() == path
    assert os.getcwd() == original_path

    # check that we return in the right path even when throwing an error
    try:
        with sim.enter_path():
            raise ValueError()
    except ValueError:
        assert os.getcwd() == original_path


@pytest.mark.parametrize(
    "array",
    [
        np.array([1, 2, 3]),
        np.array([[1, 2, 3], [2, 3, 4]]),
        np.array([1]),
        np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]),
    ],
)
def test_add_field_dimensions(temp_manager, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_field("test", array)

    sim.finish_step()
    sim.add_field("test", 2 * array)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    np.testing.assert_array_equal(sim.data["test"].at_step(0), array)
    np.testing.assert_array_equal(sim.data["test"].at_step(1), 2 * array)


def test_add_fields(temp_manager):
    # Mock the add_field method
    simulation_writer = temp_manager.create_simulation()

    # Test data
    fields = {
        "field1": np.array([1, 2, 3]),
        "field2": (np.array([4, 5, 6]), "Cell"),
        "field3": np.array([7, 8, 9]),
    }
    time = 1.0
    mesh = "test_mesh"

    # Call the method
    simulation_writer.add_fields(fields, time, mesh)

    assert simulation_writer.data.keys() == fields.keys()
    assert simulation_writer.data["field2/0"].attrs["center"] == "Cell"
    assert simulation_writer.data["field1/0"].attrs["center"] == "Node"
    np.testing.assert_array_equal(
        simulation_writer.data["field3"][0], np.array([7, 8, 9])
    )
